// Copyright 2022 Citra Emulator Project
// Licensed under GPLv2 or any later version
// Refer to the license.txt file included.

#include <glad/glad.h>
#include "common/assert.h"
#include "common/settings.h"
#include "core/telemetry_session.h"
#include "video_core/renderer_opengl/gl_driver.h"

namespace OpenGL {

DECLARE_ENUM_FLAG_OPERATORS(DriverBug);

inline std::string_view GetSource(GLenum source) {
#define RET(s)                                                                                     \
    case GL_DEBUG_SOURCE_##s:                                                                      \
        return #s
    switch (source) {
        RET(API);
        RET(WINDOW_SYSTEM);
        RET(SHADER_COMPILER);
        RET(THIRD_PARTY);
        RET(APPLICATION);
        RET(OTHER);
    default:
        UNREACHABLE();
    }
#undef RET

    return std::string_view{};
}

inline std::string_view GetType(GLenum type) {
#define RET(t)                                                                                     \
    case GL_DEBUG_TYPE_##t:                                                                        \
        return #t
    switch (type) {
        RET(ERROR);
        RET(DEPRECATED_BEHAVIOR);
        RET(UNDEFINED_BEHAVIOR);
        RET(PORTABILITY);
        RET(PERFORMANCE);
        RET(OTHER);
        RET(MARKER);
    default:
        UNREACHABLE();
    }
#undef RET

    return std::string_view{};
}

static void APIENTRY DebugHandler(GLenum source, GLenum type, GLuint id, GLenum severity,
                                  GLsizei length, const GLchar* message, const void* user_param) {
    Log::Level level = Log::Level::Info;
    switch (severity) {
    case GL_DEBUG_SEVERITY_HIGH:
        level = Log::Level::Critical;
        break;
    case GL_DEBUG_SEVERITY_MEDIUM:
        level = Log::Level::Warning;
        break;
    case GL_DEBUG_SEVERITY_NOTIFICATION:
    case GL_DEBUG_SEVERITY_LOW:
        level = Log::Level::Debug;
        break;
    }

    LOG_GENERIC(Log::Class::Render_OpenGL, level, "{} {} {}: {}", GetSource(source), GetType(type),
                id, message);
}

Driver::Driver(Core::TelemetrySession& telemetry_session_) : telemetry_session{telemetry_session_} {
    const bool enable_debug = Settings::values.renderer_debug.GetValue();
    if (enable_debug) {
        glEnable(GL_DEBUG_OUTPUT);
        glDebugMessageCallback(DebugHandler, nullptr);
    }

    ReportDriverInfo();
    DeduceVendor();
    CheckExtensionSupport();
    FindBugs();
}

Driver::~Driver() = default;

bool Driver::HasBug(DriverBug bug) const {
    return True(bugs & bug);
}

void Driver::ReportDriverInfo() {
    // Report the context version and the vendor string
    gl_version = std::string_view{reinterpret_cast<const char*>(glGetString(GL_VERSION))};
    gpu_vendor = std::string_view{reinterpret_cast<const char*>(glGetString(GL_VENDOR))};
    gpu_model = std::string_view{reinterpret_cast<const char*>(glGetString(GL_RENDERER))};

    LOG_INFO(Render_OpenGL, "GL_VERSION: {}", gl_version);
    LOG_INFO(Render_OpenGL, "GL_VENDOR: {}", gpu_vendor);
    LOG_INFO(Render_OpenGL, "GL_RENDERER: {}", gpu_model);

    // Add the information to the telemetry system
    constexpr auto user_system = Common::Telemetry::FieldType::UserSystem;
    telemetry_session.AddField(user_system, "GPU_Vendor", std::string{gpu_vendor});
    telemetry_session.AddField(user_system, "GPU_Model", std::string{gpu_model});
    telemetry_session.AddField(user_system, "GPU_OpenGL_Version", std::string{gl_version});
}

void Driver::DeduceVendor() {
    if (gpu_vendor.find("NVIDIA") != gpu_vendor.npos) {
        vendor = Vendor::Nvidia;
    } else if ((gpu_vendor.find("ATI") != gpu_vendor.npos) ||
               (gpu_vendor.find("AMD") != gpu_vendor.npos) ||
               (gpu_vendor.find("Advanced Micro Devices") != gpu_vendor.npos)) {
        vendor = Vendor::AMD;
    } else if (gpu_vendor.find("Intel") != gpu_vendor.npos) {
        vendor = Vendor::Intel;
    } else if (gpu_vendor.find("ARM") != gpu_vendor.npos) {
        vendor = Vendor::ARM;
    } else if (gpu_vendor.find("Qualcomm") != gpu_vendor.npos) {
        vendor = Vendor::Qualcomm;
    } else if (gpu_vendor.find("Samsung") != gpu_vendor.npos) {
        vendor = Vendor::Samsung;
    } else if (gpu_vendor.find("GDI Generic") != gpu_vendor.npos) {
        vendor = Vendor::Generic;
    }
}

void Driver::CheckExtensionSupport() {
    ext_buffer_storage = GLAD_GL_EXT_buffer_storage;
    arb_buffer_storage = GLAD_GL_ARB_buffer_storage;
    arb_clear_texture = GLAD_GL_ARB_clear_texture;
    arb_get_texture_sub_image = GLAD_GL_ARB_get_texture_sub_image;
    ext_clip_cull_distance = GLAD_GL_EXT_clip_cull_distance;
    is_suitable = GLAD_GL_VERSION_4_3 || GLAD_GL_ES_VERSION_3_1;
}

void Driver::FindBugs() {
#ifdef __unix__
    const bool is_linux = true;
#else
    const bool is_linux = false;
#endif

    // TODO: Check if these have been fixed in the newer driver
    if (vendor == Vendor::AMD) {
        bugs |= DriverBug::ShaderStageChangeFreeze | DriverBug::VertexArrayOutOfBound;
    }

    if (vendor == Vendor::AMD || (vendor == Vendor::Intel && !is_linux)) {
        bugs |= DriverBug::BrokenTextureView;
    }
}

} // namespace OpenGL
